;;; defcapture.el --- A convenience macro for the Doct DSL       -*- lexical-binding: t; -*-

;; Copyright (C) 2023  Abraham Aguilar

;; Author: Abraham Aguilar <a.aguilar@ciencias.unam.mx>
;; URL: https://github.com/aggu4/defcapture
;; Keywords: convenience org
;; Description: A defun analog for org-capture templates
;; Package-Version: 20230909.353
;; Package-Revision: 777a10a3343d
;; Package-Requires: ((emacs "25.1") (doct "3.0"))
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

;;; Commentary:

;; This is a simple convenience macro for defining org capture templates.
;; It uses the Doct DSL for declarations. It has a dedicated namespace
;; for capture templates and provides convenient functions to manage
;; capture templates.
;;
;; Example of usage:
;;
;; (defcapture parent () "Parent Capture"
;;  :keys "p"
;;  :file "~/example.org"
;;  :prepend t
;;  :template ("* %{todo-state} %^{Description}"
;;             ":PROPERTIES:"
;;             ":Created: %U"
;;             ":END:"
;;             "%?"))
;;
;; (defcapture childa (parent) "First Child"
;;   :keys "a"
;;   :headline "One"
;;   :todo-state "TODO"
;;   :hook (lambda () (message "\" First Child\" selected.")))
;;
;; (defcapture childb (parent) "Second Child"
;;   :keys "b"
;;   :headline "Two"
;;   :todo-state "NEXT")
;;
;; (defcapture childc (parent) "Third Child"
;;   :keys "c"
;;   :headline "Three"
;;   :todo-state "MAYBE")

;;; Code:

(require 'doct)
(require 'cl-lib)

(cl-defstruct defcapture--capture
  "A structure for captures. PARENTS is a list of symbols which are the.
names of the capture's parents. Declaration is the declaration of the
capture in Doct DSL."
  parents
  declaration)

(defvar defcapture--capture-namespace (make-hash-table)
  "The variable that holds the namespace for capture templates.")

(defun defcapture--capture-boundp (name)
  "Return t if NAME is bound to a capture and nil if it isn't."
  (or (and (gethash name defcapture--capture-namespace) t)
      (and (message "The capture %s isn't defined" name) nil)))


(defun defcapture--children (name)
  "List the children of capture NAME."
  (cl-loop for k being each hash-key of defcapture--capture-namespace
           using (hash-value v)
           appending
           (when (cl-member name (defcapture--capture-parents v))
             (list k))))

(defun defcapture--declaration (name)
  "Return the declaration for `defcapture--capture' NAME collecting it's children."
  (let ((children (defcapture--children name)))
    (if children
        (append (defcapture--capture-declaration
                 (gethash name defcapture--capture-namespace))
                (list
                 :children (mapc #'defcapture--declaration
                                 children)))
      (defcapture--capture-declaration
        (gethash name defcapture--capture-namespace)))))


;;;###autoload
(defun defcapture-captures ()
  "Sync the captures in defcapture's namespace with `org-capture-templates'."
  (doct (cl-loop for name being each hash-key of defcapture--capture-namespace
                 using (hash-value capture)
                 appending
                 (unless (defcapture--capture-parents capture)
                   (list (defcapture--declaration name))))))


;;;###autoload
(cl-defmacro defcapture (name (&rest parents) desc &body body)
  "Define a capture NAME and, optionally, declare it's PARENTS.
DESC must be a string. BODY is the capture template's declarations in
doct style."
  (declare (indent defun))
  `(if (cl-every (lambda (x) (and x t))
                 (mapc #'defcapture--capture-boundp ',parents))
       (puthash ',name
                (make-defcapture--capture
                 :parents ',parents
                 :declaration ',(cons desc body))
                defcapture--capture-namespace)
     (progn
       (message "The parent %s is not defined."
        (cl-find-if #'defcapture--capture-boundp ',parents)))))

;;;###autoload
(defun defcapture-generate-defcapture (name)
  "Return capture NAME's definition if NAME is `defcapture--capture-boundp'.
Otherwise return nil."
  (when (defcapture--capture-boundp name)
    (let ((capture (gethash name defcapture--capture-namespace)))
      `(defcapture ,name (,@(defcapture--capture-parents capture))
         ,@(defcapture--capture-declaration capture)))))


;;;###autoload
(defun defcapture-remove-capture (name)
  "Remove the capture NAME and it's children recursively."
  (when (defcapture--capture-boundp name)
    (mapc #'defcapture-remove-capture (defcapture--children name))
    (remhash name defcapture--capture-namespace)))



(provide 'defcapture)
;;; defcapture.el ends here
